#include "danzeff.h"
#include "pspkernel.h"

// headerfiles needed for the GU and VRAM versions
#include <malloc.h>
#include "pspgu.h"
#include "png.h"

#define false 0
#define true 1

// structures used for drawing the keyboard
	struct danzeff_vertex
	{
		float u, v;
		unsigned int color;
		float x,y,z;
	};

	struct danzeff_gu_surface
	{
		u32		surface_width;
		u32		surface_height;
		u32		texture_width;
		u32		texture_height;
		u32		*texture;
	};

/*bool*/ int holding = false;     //user is holding a button
/*bool*/ int dirty = true;        //keyboard needs redrawing
/*bool*/ int shifted = false;     //user is holding shift
int mode = 0;             //charset selected. (0 - letters or 1 - numbers)
/*bool*/ int initialized = false; //keyboard is initialized

//Position on the 3-3 grid the user has selected (range 0-2)
static int selected_x = 1;
static int selected_y = 1;

// location that we are moved to
int moved_x = 480 - 170;
int moved_y = 272 - 170;

//Variable describing where each of the images is
#define guiStringsSize 18 /* size of guistrings array */
#define PICS_BASEDIR "./graphics/"

static char *guiStrings[] = 
{
	"/keys.png", "/keys_t.png", "/keys_s.png",
	"/keys_c.png", "/keys_c_t.png", "/keys_s_c.png",
	"/nums.png", "/nums_t.png", "/nums_s.png",
	"/nums_c.png", "/nums_c_t.png", "/nums_s_c.png",
  "/keys2_c.png", "/keys2_c_t.png", "/keys2_s_c.png",
  "/keys3_c.png", "/keys3_c_t.png", "/keys3_s_c.png",
};

//amount of modes (non shifted), each of these should have a corresponding shifted mode.
#define MODE_COUNT 3
#define MAX_VKEYBOARD 6
//this is the layout of the keyboard
static char modeChar[MAX_VKEYBOARD][3][3][5] = 
{
	{	//standard letters
		{ ",abc",  ".def","!ghi" },
		{ "-jkl", { DANZEFF_DEL,          'm', ' ', 'n', '\0' }, "|opq" },
		{ "(rst",  ":uvw",")xyz" }
	},
	
	{	//capital letters
		{ "^ABC",  "\\DEF","*GHI" },
		{ "_JKL", { DANZEFF_DEL,    'M', ' ', 'N', '\0' }, "\"OPQ" },
		{ "=RST",  ";UVW","/XYZ" }
	},
	
	{	//numbers
		{ {   DANZEFF_CTRL_C ,   DANZEFF_F1 , 0                 , '1', '\0' }, 
      {   DANZEFF_ESCAPE ,   DANZEFF_F2 , 0                 , '2', '\0' }, 
      {   0              ,   DANZEFF_F3 , 0                 , '3', '\0' }},

		{ {   DANZEFF_TAB    ,   DANZEFF_F4 , DANZEFF_CTRL_L    , '4', '\0' }, 
      {   DANZEFF_DEL    ,   DANZEFF_F5 ,                ' ', '5', '\0' }, 
      {   DANZEFF_INSERT ,   DANZEFF_F6 , DANZEFF_CTRL_M    , '6', '\0' }},

		{ {   DANZEFF_PAGE_UP,   DANZEFF_F7 , DANZEFF_PAGE_DOWN , '7', '\0' }, 
      {   DANZEFF_HOME   ,   DANZEFF_F8 , DANZEFF_END       , '8', '\0' }, 
      {   '.'            ,   DANZEFF_F9 ,                '0', '9', '\0' }},
	},

	{	//special characters
		{ ",(.)",  "\"<'>","-[_]" },
		{ "!{?}", { DANZEFF_DEL, 0, ' ', 0 }, "+\\=/" },
		{ ":@;#",  "~$`%","*^|&" }
	},

	{	//control
		{ {   0     , DANZEFF_CTRL_A , DANZEFF_CTRL_B , DANZEFF_CTRL_C, '\0' }, 
      {   0     , DANZEFF_CTRL_D , DANZEFF_CTRL_E , DANZEFF_CTRL_F, '\0' }, 
      {   0     , DANZEFF_CTRL_G , DANZEFF_CTRL_H , DANZEFF_CTRL_I, '\0' }},

		{ {   0     , DANZEFF_CTRL_J , DANZEFF_CTRL_K , DANZEFF_CTRL_L, '\0' }, 
      {   0     , DANZEFF_CTRL_M ,             ' ', DANZEFF_CTRL_N, '\0' }, 
      {   0     , DANZEFF_CTRL_O , DANZEFF_CTRL_P , DANZEFF_CTRL_Q, '\0' }},

		{ {   0     , DANZEFF_CTRL_R , DANZEFF_CTRL_S , DANZEFF_CTRL_T, '\0' }, 
      {   0     , DANZEFF_CTRL_U , DANZEFF_CTRL_V , DANZEFF_CTRL_W, '\0' }, 
      {   0     , DANZEFF_CTRL_X , DANZEFF_CTRL_Y , DANZEFF_CTRL_Z, '\0' }},
  },

	{	//alt
		{ {   0     , DANZEFF_ALT_A , DANZEFF_ALT_B , DANZEFF_ALT_C, '\0' }, 
      {   0     , DANZEFF_ALT_D , DANZEFF_ALT_E , DANZEFF_ALT_F, '\0' }, 
      {   0     , DANZEFF_ALT_G , DANZEFF_ALT_H , DANZEFF_ALT_I, '\0' }},

		{ {   0     , DANZEFF_ALT_J , DANZEFF_ALT_K , DANZEFF_ALT_L, '\0' }, 
      {   0     , DANZEFF_ALT_M ,             ' ', DANZEFF_ALT_N, '\0' }, 
      {   0     , DANZEFF_ALT_O , DANZEFF_ALT_P , DANZEFF_ALT_Q, '\0' }},

		{ {   0     , DANZEFF_ALT_R , DANZEFF_ALT_S , DANZEFF_ALT_T, '\0' }, 
      {   0     , DANZEFF_ALT_U , DANZEFF_ALT_V , DANZEFF_ALT_W, '\0' }, 
      {   0     , DANZEFF_ALT_X , DANZEFF_ALT_Y , DANZEFF_ALT_Z, '\0' }},
	}
};

int 
danzeff_isinitialized()
{
	return initialized;
}

int 
danzeff_dirty()
{
	return dirty;
}

/** Attempts to read a character from the controller
* If no character is pressed then we return 0
* Other special values: 1 = move left, 2 = move right, 3 = select, 4 = start
* Every other value should be a standard ascii value.
* An unsigned int is returned so in the future we can support unicode input
*/
unsigned int 
danzeff_readInput(SceCtrlData pspctrl)
{
	//Work out where the analog stick is selecting
	int x = 1;
	int y = 1;
	if (pspctrl.Lx < 85)     x -= 1;
	else if (pspctrl.Lx > 170) x += 1;
	if (pspctrl.Ly < 85)     y -= 1;
	else if (pspctrl.Ly > 170) y += 1;
	if (selected_x != x || selected_y != y) //If they've moved, update dirty
	{
		dirty = true;
		selected_x = x;
		selected_y = y;
	}
	//if they are changing shift then that makes it dirty too
	if ((!shifted && (pspctrl.Buttons & PSP_CTRL_RTRIGGER)) || (shifted && !(pspctrl.Buttons & PSP_CTRL_RTRIGGER)))
		dirty = true;

	unsigned int pressed = 0; //character they have entered, 0 as that means 'nothing'
	shifted = (pspctrl.Buttons & PSP_CTRL_RTRIGGER)?true:false;

	if (!holding)
	{
		if (pspctrl.Buttons& (PSP_CTRL_CROSS|PSP_CTRL_CIRCLE|PSP_CTRL_TRIANGLE|PSP_CTRL_SQUARE)) //pressing a char select button
		{
			int innerChoice = 0;
			if      (pspctrl.Buttons & PSP_CTRL_TRIANGLE)
				innerChoice = 0;
			else if (pspctrl.Buttons & PSP_CTRL_SQUARE)
				innerChoice = 1;
			else if (pspctrl.Buttons & PSP_CTRL_CROSS)
				innerChoice = 2;
			else //if (pspctrl.Buttons & PSP_CTRL_CIRCLE)
				innerChoice = 3;

			//Now grab the value out of the array
			pressed = modeChar[ (mode*2 + shifted) % MAX_VKEYBOARD][y][x][innerChoice];
		}
		else if (pspctrl.Buttons& PSP_CTRL_LTRIGGER) //toggle mode
		{
			dirty = true;
			mode++;
			mode %= MODE_COUNT;
		}
		else if (pspctrl.Buttons& PSP_CTRL_LEFT)
		{
			pressed = DANZEFF_LEFT;
		}
		else if (pspctrl.Buttons& PSP_CTRL_RIGHT)
		{
			pressed = DANZEFF_RIGHT;
		}
		else if (pspctrl.Buttons& PSP_CTRL_DOWN)
		{
			pressed = DANZEFF_DOWN;
		}
		else if (pspctrl.Buttons& PSP_CTRL_UP)
		{
			pressed = DANZEFF_UP;
		}
		else if (pspctrl.Buttons& PSP_CTRL_SELECT)
		{
			pressed = DANZEFF_SELECT; //SELECT
		}
		else if (pspctrl.Buttons& PSP_CTRL_START)
		{
			pressed = DANZEFF_START; //START
		}
	}

	holding = pspctrl.Buttons & ~PSP_CTRL_RTRIGGER; //RTRIGGER doesn't set holding

	return pressed;
}

///-------------------------------------------------------------------------------
///These are specific to the implementation, they should have the same behaviour across implementations.
///-------------------------------------------------------------------------------


struct danzeff_gu_surface	keyTextures[guiStringsSize];



///Internal function to draw a surface internally offset
//Render the given surface at the current screen position offset by screenX, screenY
//the surface will be internally offset by offsetX,offsetY. And the size of it to be drawn will be intWidth,intHeight
void surface_draw_offset(struct danzeff_gu_surface* surface, int screenX, int screenY, int offsetX, int offsetY, int intWidth, int intHeight, int transperant)
{
	sceGuEnable( GU_TEXTURE_2D );
	sceGuAlphaFunc( GU_GREATER, 0, 0xff );
	sceGuEnable( GU_ALPHA_TEST );
	sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);
	sceGuTexEnvColor(0xFF000000);
	sceGuBlendFunc( GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0 );
	sceGuEnable( GU_BLEND );
	sceGuTexMode(GU_PSM_8888,0,0,GU_FALSE);
	sceGuTexImage(0,surface->surface_width, surface->surface_height,surface->surface_width, surface->texture);
	sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);

	struct danzeff_vertex* c_vertices = (struct danzeff_vertex*)sceGuGetMemory(6 * sizeof(struct danzeff_vertex));

	c_vertices[0].u 		= offsetX;
	c_vertices[0].v 		= offsetY;
	c_vertices[0].x 		= moved_x + screenX;
	c_vertices[0].y 		= moved_y + screenY;
	c_vertices[0].z 		= 0;

	c_vertices[1].u 		= offsetX + intWidth;
	c_vertices[1].v 		= offsetY;
	c_vertices[1].x 		= moved_x + screenX + intWidth;
	c_vertices[1].y 		= moved_y + screenY;
	c_vertices[1].z 		= 0;

	c_vertices[2].u 		= offsetX;
	c_vertices[2].v 		= offsetY + intHeight;
	c_vertices[2].x 		= moved_x + screenX;
	c_vertices[2].y 		= moved_y + screenY + intHeight;
	c_vertices[2].z 		= 0;

	memcpy(&c_vertices[3], &c_vertices[2], sizeof(struct danzeff_vertex));
	memcpy(&c_vertices[4], &c_vertices[1], sizeof(struct danzeff_vertex));

	c_vertices[5].u 		= offsetX + intWidth;
	c_vertices[5].v 		= offsetY + intHeight;
	c_vertices[5].x 		= moved_x + screenX + intWidth;
	c_vertices[5].y 		= moved_y + screenY + intHeight;
	c_vertices[5].z 		= 0;

	if (transperant == true)
		{
		c_vertices[0].color 	= 0x80FFFFFF;
		c_vertices[1].color 	= 0x80FFFFFF;
		c_vertices[2].color 	= 0x80FFFFFF;
		c_vertices[3].color 	= 0x80FFFFFF;
		c_vertices[4].color 	= 0x80FFFFFF;
		c_vertices[5].color 	= 0x80FFFFFF;
		}
	else
		{
		c_vertices[0].color 	= 0xFFFFFFFF;
		c_vertices[1].color 	= 0xFFFFFFFF;
		c_vertices[2].color 	= 0xFFFFFFFF;
		c_vertices[3].color 	= 0xFFFFFFFF;
		c_vertices[4].color 	= 0xFFFFFFFF;
		c_vertices[5].color 	= 0xFFFFFFFF;
		}

	sceGuDrawArray(GU_TRIANGLES,GU_TEXTURE_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_TRANSFORM_2D, 6, 0, c_vertices);

	sceGuDisable( GU_BLEND );
	sceGuDisable( GU_ALPHA_TEST );
	sceGuDisable( GU_TEXTURE_2D );
}

void danzeff_block_copy(struct danzeff_gu_surface* surface, u32 *texture)
{
	u32 *dest = surface->texture;
	u32 stride = surface->surface_width - surface->texture_width;
	u32 y, x;

	for (y = 0 ; y < surface->texture_height ; y++)
	{
		for (x = 0 ; x < surface->texture_width ; x++)
		{
		*dest++ = *texture++;
		}
		// skip at the end of each line
		if (stride > 0)
		{
			dest += stride;
		}
	}
}

static void danzeff_user_warning_fn(png_structp png_ptr, png_const_charp warning_msg)
{
	// ignore PNG warnings
}

/* Get the width and height of a png image */
int danzeff_get_png_image_size(const char* filename, u32 *png_width, u32 *png_height)
{
	png_structp png_ptr;
	png_infop info_ptr;
	unsigned int sig_read = 0;
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;
	FILE *fp;

	if ((fp = fopen(filename, "rb")) == NULL) return -1;
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		fclose(fp);
		return -1;
	}
	png_set_error_fn(png_ptr, (png_voidp) NULL, (png_error_ptr) NULL, danzeff_user_warning_fn);
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
		return -1;
	}
	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, sig_read);
	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, int_p_NULL, int_p_NULL);

	png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	fclose(fp);

	*png_width = width;
	*png_height = height;
	return 0;
}

/* Load a texture from a png image */
int danzeff_load_png_image(const char* filename, u32 *ImageBuffer)
{
	png_structp png_ptr;
	png_infop info_ptr;
	unsigned int sig_read = 0;
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;
	size_t x, y;
	u32* line;
	FILE *fp;

	if ((fp = fopen(filename, "rb")) == NULL) return -1;
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL) {
		fclose(fp);
		return -1;
	}
	png_set_error_fn(png_ptr, (png_voidp) NULL, (png_error_ptr) NULL, danzeff_user_warning_fn);
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
		return -1;
	}
	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, sig_read);
	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, int_p_NULL, int_p_NULL);
	png_set_strip_16(png_ptr);
	png_set_packing(png_ptr);
	if (color_type == PNG_COLOR_TYPE_PALETTE) png_set_palette_to_rgb(png_ptr);
	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8) png_set_gray_1_2_4_to_8(png_ptr);
	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) png_set_tRNS_to_alpha(png_ptr);
	png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
	line = (u32*) malloc(width * 4);
	if (!line) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
		return -1;
	}
	for (y = 0; y < height; y++)
	{
		png_read_row(png_ptr, (u8*) line, png_bytep_NULL);
		for (x = 0; x < width; x++)
		{
			ImageBuffer[y*width+x] = line[x];
		}
	}
	free(line);
	png_read_end(png_ptr, info_ptr);
	png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	fclose(fp);
	return 0;
}

u32 danzeff_convert_pow2(u32 size)
{
	u32 pow_counter = 0;

	for ( ; pow_counter < 32 ; pow_counter++)
	{
		// Find the first value which is higher
		if ((size >> pow_counter) == 0)
		{
			// take already good values into account
			if ((1 << pow_counter) != size)
			{
				return (1 << pow_counter);
			}
			else
			{
				return (1 << (pow_counter-1));
			}
		}
	}
	return 0;
}

/* load all the guibits that make up the OSK */
int 
danzeff_load()
{
  char tmp_filename[128];
	u32 *temp_texture;

	if (initialized) return 1;

	int a;
	for (a = 0; a < guiStringsSize; a++)
	{
		u32 height, width;
    strcpy(tmp_filename, PICS_BASEDIR);
    strcat(tmp_filename, guiStrings[a] );

		if (!((a-1)%3)) //skip loading the _t files
		{
			keyTextures[a].texture = NULL;
			continue;
		}
		if (danzeff_get_png_image_size(tmp_filename, &width, &height) == 0)
		{
			// The png is always converted to PSM_8888 format when read
			temp_texture = (u32 *)malloc(width*height*4);
			if (danzeff_load_png_image(tmp_filename, temp_texture) != 0)
			{
        fprintf(stdout, "failed to load %s\n", tmp_filename);
				// Error .. Couldn't get png info from one of the needed files
				int b;

				for (b = 0; b < a; b++)
				{
					free(keyTextures[b].texture);
					keyTextures[b].texture = NULL;
				}
				initialized = false;
				return 0;
			}
			else
			{
				// we need to store the texture in an image of width and heights of 2^n sizes
				keyTextures[a].texture_width 	= (float)width;
				keyTextures[a].texture_height 	= (float)height;
				keyTextures[a].surface_width 	= (float)danzeff_convert_pow2(width);
				keyTextures[a].surface_height 	= (float)danzeff_convert_pow2(height);
				keyTextures[a].texture 			= (u32 *)malloc(keyTextures[a].surface_width*keyTextures[a].surface_height*4);
				// block copy the texture into the surface
				danzeff_block_copy(&keyTextures[a], temp_texture);
				free(temp_texture);
			}
		}
		else
		{
      fprintf(stdout, "failed to load %s\n", tmp_filename);
			// Error .. Couldn't get png info from one of the needed files
			int b;

			for (b = 0; b < a; b++)
			{
				free(keyTextures[b].texture);
				keyTextures[b].texture = NULL;
			}
			initialized = false;
			return 0;
		}
	}
	initialized = true;
  return 1;
}
/* remove all the guibits from memory */
void danzeff_free()
{
	if (!initialized) return;

	int a;
	for (a = 0; a < guiStringsSize; a++)
	{
		free(keyTextures[a].texture);
		keyTextures[a].texture = NULL;
	}
	initialized = false;
}

/* draw the keyboard at the current position */
void danzeff_render()
{
	int	transperant;
	int	x = selected_x*64;
	int y = selected_y*64;
	dirty = false;

	///Draw the background for the selected keyboard either transparent or opaque
	///this is the whole background image, not including the special highlighted area
	//if center is selected then draw the whole thing opaque
	if (selected_x == 1 && selected_y == 1)
		{
		transperant = false;
		}
	else
		{
		transperant = true;
		}

  int index = (mode*2 + shifted) % MAX_VKEYBOARD;
	surface_draw_offset(&keyTextures[index*3], 0, 0, 0, 0, keyTextures[index*3].texture_width,
																	 	 keyTextures[index*3].texture_height,
																	 	 transperant);

	///Draw the current Highlighted Selector (orange bit) based on the orientation

	surface_draw_offset(&keyTextures[index*3 + 2],
	//Offset from the current draw position to render at
	selected_x*43, selected_y*43,
	//internal offset of the image
	x,y,
	//size to render (always the same)
	64, 64,
	false);
}

/* move the position the keyboard is currently drawn at */
void danzeff_moveTo(const int newX, const int newY)
{
	moved_x = newX;
	moved_y = newY;
	dirty = true;
}
